package ru.dbotthepony.kommons.event

import java.util.function.Consumer

interface ISubscriptable<V> {
	/**
	 * Listener token, allows to remove listener from subscriber list
	 */
	fun interface L {
		/**
		 * Removes this listener
		 */
		fun remove()
	}

	fun addListener(listener: Consumer<V>): L

	class Impl<V> : ISubscriptable<V>, Consumer<V> {
		private inner class L(val callback: Consumer<V>) : ISubscriptable.L {
			private var isRemoved = false

			init {
				subscribers.add(this)
			}

			override fun remove() {
				if (!isRemoved) {
					isRemoved = true
					queue.add(this)
				}
			}
		}

		private val subscribers = ArrayList<L>(0)
		private val queue = ArrayList<L>(0)

		override fun addListener(listener: Consumer<V>): ISubscriptable.L {
			return L(listener)
		}

		override fun accept(t: V) {
			queue.forEach { subscribers.remove(it) }
			queue.clear()
			subscribers.forEach { it.callback.accept(t) }
		}
	}

	companion object : ISubscriptable<Nothing>, L {
		@Suppress("unchecked_cast")
		fun <T> empty(): ISubscriptable<T> {
			return this as ISubscriptable<T>
		}

		override fun remove() {}

		override fun addListener(listener: Consumer<Nothing>): L {
			return this
		}
	}
}
