package ru.dbotthepony.kommons.io

import java.io.DataInputStream
import java.io.DataOutputStream

class EnumValueCodec<V : Enum<V>>(clazz: Class<out V>) : IStreamCodec<V> {
	val clazz = searchClass(clazz)
	val values: List<V> = listOf(*this.clazz.enumConstants!!)
	val valuesMap = values.associateBy { it.name }

	override fun read(stream: DataInputStream): V {
		val id = stream.readVarInt()
		return values.getOrNull(id) ?: throw NoSuchElementException("No such enum with index $id")
	}

	override fun write(stream: DataOutputStream, value: V) {
		stream.writeVarInt(value.ordinal)
	}

	override fun copy(value: V): V {
		return value
	}

	override fun compare(a: V, b: V): Boolean {
		return a === b
	}

	companion object {
		/**
		 * FIXME: enums with abstract methods which get compiled to subclasses, whose DO NOT expose "parent's" enum constants array
		 *
		 * is there an already existing solution?
		 */
		fun <V : Enum<V>> searchClass(clazz: Class<out V>): Class<out V> {
			var search: Class<*> = clazz

			while (search.enumConstants == null && search.superclass != null) {
				search = search.superclass
			}

			if (search.enumConstants == null) {
				throw ClassCastException("$clazz does not represent an enum or enum subclass")
			}

			return search as Class<out V>
		}
	}
}