package ru.dbotthepony.kommons.io

import java.io.DataOutput
import java.io.OutputStream
import java.io.RandomAccessFile
import java.math.BigDecimal
import java.util.function.IntConsumer

fun OutputStream.writeBigDecimal(value: BigDecimal) {
	writeInt(value.scale())
	val bytes = value.unscaledValue().toByteArray()
	writeVarInt(bytes.size)
	write(bytes)
}

fun <S : OutputStream, V> S.writeCollection(collection: Collection<V>, writer: S.(V) -> Unit) {
	writeVarInt(collection.size)

	for (value in collection) {
		writer(this, value)
	}
}

fun OutputStream.writeInt(value: Int) {
	if (this is DataOutput) {
		writeInt(value)
		return
	}

	write(value ushr 24)
	write(value ushr 16)
	write(value ushr 8)
	write(value)
}

fun OutputStream.writeLong(value: Long) {
	if (this is DataOutput) {
		writeLong(value)
		return
	}

	write((value ushr 48).toInt())
	write((value ushr 40).toInt())
	write((value ushr 32).toInt())
	write((value ushr 24).toInt())
	write((value ushr 16).toInt())
	write((value ushr 8).toInt())
	write(value.toInt())
}

fun OutputStream.writeFloat(value: Float) = writeInt(value.toBits())
fun OutputStream.writeDouble(value: Double) = writeLong(value.toBits())

private fun writeVarLong(write: IntConsumer, value: Long): Int {
	var value = value
	var i = 0

	do {
		val toWrite = (value and 0x7F).toInt()
		value = value ushr 7

		if (value == 0L)
			write.accept(toWrite)
		else
			write.accept(toWrite or 0x80)

		i++
	} while (value != 0L)

	return i
}

private fun writeVarInt(write: IntConsumer, value: Int): Int {
	return writeVarLong(write, value.toLong())
}

private fun Int.toSignedVar(): Int {
	if (this >= 0)
		return this shl 1
	else
		return (this shl 1) or 1
}

private fun Long.toSignedVar(): Long {
	if (this >= 0)
		return this shl 1
	else
		return ((-this - 1) shl 1) or 1L
}

fun RandomAccessFile.writeVarLong(value: Long) = writeVarLong(::write, value)
fun RandomAccessFile.writeVarInt(value: Int) = writeVarInt(::write, value)
fun OutputStream.writeVarLong(value: Long) = writeVarLong(::write, value)
fun OutputStream.writeVarInt(value: Int) = writeVarInt(::write, value)

fun RandomAccessFile.writeSignedVarLong(value: Long) = writeVarLong(::write, value.toSignedVar())
fun RandomAccessFile.writeSignedVarInt(value: Int) = writeVarInt(::write, value.toSignedVar())
fun OutputStream.writeSignedVarLong(value: Long) = writeVarLong(::write, value.toSignedVar())
fun OutputStream.writeSignedVarInt(value: Int) = writeVarInt(::write, value.toSignedVar())

fun OutputStream.writeBinaryString(input: String) {
	val bytes = input.encodeToByteArray()
	writeVarInt(bytes.size)
	write(bytes)
}
